'use strict';

const Excel = require('exceljs');
const Clone = require('clone');
const urlMaster = require('locutus/php/url');
const dateMaster = require('locutus/php/datetime');
const uuid = require('uuid');
const fs = require('fs');
const path = require('path');
const superagent = require('superagent');

class DataLoader {
  constructor(config, workbook) {
    this.config = config;
    this.workbook = workbook;
    this.setting = {
    };
    this.events = {
    };
    this.promise = new Promise(resolve => {
      const datas = config.datas;
      const total = config.total || 0;
      let singleMaxLimit = config.singleMaxLimit || 20000;
      const maxQuery = config.maxQuery || 20;
      let limit = 0;
      if (total > singleMaxLimit) {
        const pageSize = Math.ceil(total / singleMaxLimit);
        if (pageSize > maxQuery) {
          singleMaxLimit = Math.ceil(total / maxQuery);
        }
      }
      limit = singleMaxLimit;
      resolve(this.load(config, workbook, datas, total, limit, 0));
    });
  }

  emit(event, data) {
    if (this.events[event]) {
      this.events[event].call(this, data);
    }
  }

  on(event, fn) {
    this.events[event] = fn;
  }

  load(config, workbook, url, total, limit, offset) {
    if (typeof (url) === 'string') {
      return new Promise((resolve, reject) => {
        try {
          const parse = urlMaster.parse_url(url);
          const param = getQueryParam(url);
          param.limit = limit;
          param.offset = offset;
          const apis = [];
          apis.push(parse.scheme + '://');
          apis.push(parse.host + ':');
          apis.push(parse.port);
          apis.push(parse.path);
          apis.push('?');
          apis.push(urlMaster.http_build_query(param));
          const api = apis.join('');
          const data = {};
          superagent
            .get(api)
            .type('form')
            .send(data)
            .end((pre, response) => {
              try {
                const json = JSON.parse(response.text);
                let datas = [];
                if (config.dataParser) {
                  datas = config.dataParser(json);
                } else {
                  datas = json.data && json.data[config.field] ? json.data[config.field] : [];
                }
                const queryTotal = json.data.pages ? json.data.pages.totalCount : datas.length;
                if (total == 0) {
                  total = queryTotal;
                  this.setting.total = total;
                  this.emit('total', total);
                }
                const handle = (offset + limit);

                const end = (handle >= total || queryTotal < limit);
                let progress = 0;
                if (handle && total) {
                  progress = (Number)((handle / total) * 100).toFixed(2);
                  progress = progress >= 100 ? 100 : progress;
                }
                this.emit('progress', {
                  progress,
                  handle,
                  total
                });
                this.emit('response', json);
                this.push(config, workbook, datas);
                if (end) {
                  this.emit('done');
                  resolve({
                    config,
                    workbook,
                  });
                } else {
                  offset += limit;
                  resolve(this.load(config, workbook, url, total, limit, offset));
                }
              } catch (e) {
                e.message = 'request: ' + api + ' & error: ' + e.message;
                reject(e);
              }
              // res.end(response.text);
            });
        } catch (e) {
          reject(e);
        }
      });
    }
    return this.push(config, workbook, url);
  }

  push(config, workbook, datas) {
    const worksheet = workbook.getWorksheet(config.sheet_name);
    const headers = config.headers || [];
    // const maxLength = datas.length;
    // const unNumber = /\d+/g;

    for (let i = 0; i < datas.length; i++) {
      const node = datas[i];
      const data = [];
      const emptys = [];
      const names = [];
      const links = [];
      headers.forEach((target, index) => {
        const id = target.key;
        names.push(id);
        let value = node[id];
        if (target.formatter) {
          if (typeof (target.formatter) === 'function') {
            value = target.formatter(value, node);
          }
        }
        const isEmpty = (!value || value == '' || value == null || value.length <= 0);
        if (isEmpty && target.empty) {
          value = target.empty;
          emptys.push(index);
        }
        if (target.link) {
          const link_sheet_name = target.link.sheet;
          const link_data = workbook.worksheetData[link_sheet_name];
          if (link_data) {
            const link_targets = target.link.targets(node, link_data);
            const link_target = target.link.link(node, link_targets);
            if (link_target) {
              links.push({
                index,
                sheet: link_sheet_name,
                key: link_target._map[target.link.key],
              });
            }
          }
        }
        const length = strlen(value);
        data[index] = value == null ? '' : value;
        const haeder_max = target.maxLength || 0;
        if (length > haeder_max) {
          target.maxLength = length;
          worksheet.getColumn(target.columnIndex).width = length + 2.62;
        }
      });
      const row = worksheet.addRow(data);
      const original = workbook.worksheetData[config.sheet_name];
      if (!original) {
        workbook.worksheetData[config.sheet_name] = [];
      }
      row.eachCell((cell, colNumber) => {
        const name = names[colNumber - 1];
        const address = cell.address;
        if (!node._map) {
          node._map = {};
        }
        node._map[name] = address;
      });
      workbook.worksheetData[config.sheet_name].push(node);
      if (emptys.length > 0) {
        emptys.forEach(node => {
          row.getCell(node + 1).style.font = {
            color: {
              argb: 'FFE4E2E1',
            },
          };
        });
      }
      if (links.length > 0) {
        links.forEach(node => {
          const cell = row.getCell(node.index + 1);
          cell.style.font = {
            color: {
              argb: 'FF1278EA',
            },
          };
          cell.value = {
            formula: 'HYPERLINK("#' + node.sheet + '!' + node.key + '","1")',
            result: cell.value,
          };
        });
      }
      row.alignment = {
        vertical: 'middle',
        horizontal: 'center',
      };
      if (config.isStream) {
        setTimeout(() => {
          this.emit('commit-row', row);
          try {
            row.commit();
          } catch (e) {
          }
        }, 50);
      }

      // datas.splice(0, 1);
      // datas.length++;
    }
    datas = null;
    return true;
  }

}

class Exporter {

  constructor(config) {
    const default_config = {
      groups: {},
      headers: [],
      datas: [],
      frozen: {
        x: true,
        y: true,
      },
      filename: '',
      title: {
        show: false,
        name: '',
      },
      time: {
        show: false,
        time: '',
      },
      dataParser: null
    };
    config = config || {};
    this.config = Clone(Object.assign(config, default_config));
    this.awaitJoinList = [];
  }

  set(key, value) {
    this.config[key] = value;
    return this;
  }

  showTitle(title) {
    this.config.title.show = true;
    this.config.title.name = title;
    return this;
  }

  showTime(time) {
    this.config.time.show = true;
    this.config.time.name = time ? time : dateMaster.date('Y-m-d H:i:s');
    return this;
  }

  setHeaders(headers) {
    return this.set('headers', headers);
  }
  setGroups(groups) {
    return this.set('groups', groups);
  }

  setField(field) {
    this.config.field = field;
    return this;
  }

  setDatas(datas) {
    this.config.datas = datas;
    return this;
  }

  setDataParser(fn) {
    this.config.dataParser = fn;
    return this;
  }

  setName(name) {
    this.config.sheet_name = name;
    return this;
  }

  isBig() {
    return this.set('isStream', true);
  }

  filename(filename) {
    return this.set('filename', filename);
  }

  load(workbook) {
    const config = this.config;
    if (config.isStream) {
      config.streamTarget = uuid.v4();
    }
    if (!config.filename) {
      config.filename = config.title.name;
    }
    if (!config.sheet_name) {
      config.sheet_name = '导出数据';
    }
    if (!workbook) {
      workbook = this.create(config);
    }
    this.createWorksheet(config, workbook);
    this.workbook = workbook;
    return new DataLoader(config, workbook);
  }

  join(exporter) {
    if (!exporter) {
      return this;
    }
    this.awaitJoinList.push(exporter);
    return this;
  }

  loadJoin(workbook) {
    if (!this.awaitJoinList || this.awaitJoinList.length <= 0) {
      return Promise.resolve([]);
    }
    const awaits = [];
    this.awaitJoinList.forEach(exporter => {
      awaits.push(exporter.load(workbook));
    });
    return Promise.all(awaits);
  }

  export() {
    const loader = this.load();
    loader.export = new Promise((resolve, reject) => {
      loader.promise.then(loaded => {
        const {
          config,
          workbook,
        } = loaded;
        // const worksheet = workbook.getWorksheet(1);
        if (config.isStream) {
          return new Promise(r => {
            setTimeout(() => {
              r(workbook.commit().then(() => {
                return {
                  type: 'file',
                  data: '/files/cache/' + config.streamTarget + '/' + config.filename + '.xlsx',
                };
              }));
            }, 100);
          });
        }
        return workbook.xlsx.writeBuffer().then(buffer => {
          return {
            type: 'buffer',
            data: buffer,
          };
        });
      }).then(data => {
        resolve(data);
      }).catch(e => {
        reject(e);
      });
    });
    return loader;
  }

  createBasicSheet(config, workbook, views, real) {
    views = views || [];
    const sheet_name = config.sheet_name || '导出数据';
    const headers = config.headers || [];
    const groups = Object.keys(config.groups) == 0 ? null : config.groups;
    const datas = config.datas || [];
    const frozen = config.frozen || {};
    const mainTitle = config.title || {
      show: false,
      showTime: false,
      name: '导出数据',
    };
    const mainTime = config.time || {
      show: false,
    };
    mainTime.time = mainTime.time || new Date().toLocaleString();
    let line = 0;
    const worksheet = workbook.addWorksheet(sheet_name, {
      views,
    });
      // let sec = workbook.addWorksheet('1.1', {
      // 	views: [{
      // 		visibility: 'unvisible'
      // 	}]
      // });

    if (mainTitle.show) {
      line++;
      const titleRow = worksheet.addRow([ mainTitle.name ]);
      titleRow.height = 30;
      titleRow.alignment = {
        vertical: 'middle',
        horizontal: 'center',
      };
      titleRow.font = {
        name: 'Microsoft Yahei',
        family: 4,
        size: 16,
        bold: false,
      };
    }
    if (mainTime.show) {
      line++;
      worksheet.addRow([ '数据统计时间：' + mainTime.time ]);
    }
    const tops = groups;
    if (tops) {
      line++;
      const topRow = [];
      for (const k in tops) {
        topRow.push(tops[k]);
      }
      worksheet.addRow(topRow);
    }
    const row = [];
    let columnTarget = 0;
    headers.forEach(node => {
      ++columnTarget;
      node.columnIndex = columnTarget;
      row.push(node.name);
    });
    const unNumber = /\d+/g;
    const columnRow = worksheet.addRow(row);
    columnRow.alignment = {
      vertical: 'middle',
      horizontal: 'center',
    };
    let xSplit = 0;
    line++;

    config.columnIndex = line;

    columnRow.eachCell(function(cell, colNumber) {
      const head = headers.find(function(data) {
        return data.columnIndex == colNumber;
      });
      const column = cell.address.replace(unNumber, '');
      if (head) {
        head.column = column;
        head.address = cell.address;
        const length = strlen(head.name);
        head.maxLength = length;
        worksheet.getColumn(head.columnIndex).width = length + 2.62;
        if (head.frozen) {
          xSplit = colNumber;
        }
      }
    });

    xSplit = frozen.x ? xSplit : 0;
    const ySplit = frozen.y ? line : 0;
    const finalViews = [{
      state: 'frozen',
      xSplit,
      ySplit,
    }];
    const dataTop = line - 1;
    for (const k in tops) {
      const groups = headers.filter(data => {
        return data.group == k;
      });
      if (groups && groups.length > 0) {
        const begin = groups[0].address.replace(unNumber, '');
        const end = groups[groups.length - 1].address.replace(unNumber, '');
        worksheet.mergeCells(begin + dataTop + ':' + end + dataTop);
        const cell = worksheet.getCell(begin + dataTop);
        cell.value = tops[k];
        cell.alignment = {
          vertical: 'middle',
          horizontal: 'center',
        };
      }
    }
    const total_begin = headers[0].address.replace(unNumber, '');
    const total_end = headers[headers.length - 1].address.replace(unNumber, '');
    let titleTop = 0;
    if (mainTitle.show) {
      titleTop = 1;
      const title_merge = total_begin + titleTop + ':' + total_end + titleTop;
      worksheet.mergeCells(title_merge);
    }
    if (mainTime.show) {
      titleTop++;
      const date_merge = total_begin + titleTop + ':' + total_end + titleTop;
      worksheet.mergeCells(date_merge);
    }
    // worksheet.getCell('A3').value = {
    //     formula: 'HYPERLINK("#1.1!c1","1")',
    //     result: '1.1'
    // };
    return real ? worksheet : finalViews;
  }

  createWorkbook(config) {
    let workbook;
    if (config.isStream) {
      mkdirsSync('./public/files/cache/' + config.streamTarget);
      const options = {
        filename: './public/files/cache/' + config.streamTarget + '/' + config.filename + '.xlsx',
        useStyles: true,
        useSharedStrings: true,
      };
      workbook = new Excel.stream.xlsx.WorkbookWriter(options);
    } else {
      workbook = new Excel.Workbook();
    }

    workbook.creator = 'Ewsedu';
    workbook.lastModifiedBy = 'Ewsedu';
    workbook.created = new Date();
    workbook.modified = new Date();
    workbook.worksheetData = {};
    return workbook;
  }

  createWorksheet(config, workbook) {
    workbook = workbook || this.createWorkbook(config);
    config = config || {};
    const headers = config.headers || [];
    const groups = config.groups || null;
    const tops = groups;
    const datas = config.datas || [];
    const frozen = config.frozen || {};
    const mainTitle = config.title || {
      show: false,
      showTime: false,
      name: '导出数据',
    };
    const mainTime = config.time || {
      show: false,
    };
    mainTime.time = mainTime.time || new Date().toLocaleString();
    const line = 0;
    const unNumber = /\d+/g;

    const headbook = new Excel.Workbook();
    const views = this.createBasicSheet(config, headbook, null, false);
    const worksheet = this.createBasicSheet(config, workbook, views, true);
    return worksheet;
  }

  create(config) {
    const workbook = this.createWorkbook(config);
    return workbook;
  }
}


function getQueryParam(url, key) {
  const parse = urlMaster.parse_url(url);
  const query = parse.query || '';
  const queries = query.split('&');
  const output = {};
  queries.forEach(node => {
    const info = node.split('=');
    if (info[0]) {
      output[info[0]] = info[1];
    }
  });
  return key ? output[key] : output;
}
function strlen(str) {
  if (str == null) return 0;
  if (typeof str !== 'string') {
    str += '';
  }
  return str.replace(/[^\x00-\xff]/g, '01').length;
}

function mkdirsSync(dirname, mode) {
  if (fs.existsSync(dirname)) {
    return true;
  }
  if (mkdirsSync(path.dirname(dirname), mode)) {
    fs.mkdirSync(dirname, mode);
    return true;
  }
}

module.exports = Exporter;
