'use strict';
const Excel = require('exceljs');
const fs = require('fs');

class Reader {

  constructor(filepath = '') {
    this._filepath = filepath;
    this._workbook = new Excel.Workbook();
  }

  get filepath() {
    return this._filepath;
  }

  get workbook() {
    return this._workbook;
  }

  readFile() {
    return new Promise(resolve => {
      const stream = fs.createReadStream(this.filepath);
      const input = this.workbook.xlsx.createInputStream();
      stream.pipe(input);
      input.on('done', () => {
        resolve(this.workbook);
      }).on('error', () => {
        resolve(this.workbook);
      });
    });
  }

  overwrite(path = this.filepath) {
    return new Promise(resolve => {
      this._workbook.xlsx.writeBuffer().then(buffer => {
        const writeStream = fs.createWriteStream(path);
        writeStream.on('finish', () => {
          resolve();
        });
        writeStream.on('error', () => {
          resolve();
        });
        writeStream.write(buffer);
        writeStream.end();
      });
    });
  }

  read() {
    return this.readFile().then(workbook => {
      const worksheet = workbook.getWorksheet(1);
      const maps = [];
      const datas = [];
      const row = worksheet.getRow(1);
      row.values.forEach(value => {
        maps.push(value);
      });
      worksheet.eachRow((row, rowNumber) => {
        if (rowNumber > 1) {
          const data = {};
          const values = row.values || [];
          maps.forEach((k, i) => {
            let value = values[i + 1] || '';
            let text = '';
            if (value instanceof Object) {
              if (value.text) {
                value = value.text;
              }
              if (value.richText instanceof Array) {
                value.richText.forEach(rich => {
                  text += rich.text;
                });
                value = text;
              }
            }
            value = value || '';
            try {
              value = value.toString();
              value = value.replace(/^\s+|\s+$/gm, '');
            } catch (e) {
              value = (value || '');
            }
            if (k) {
              data[k] = value;
            }
          });
          datas.push(data);
        }
      });
      return {
        maps,
        datas,
        workbook
      };
    }).catch(e => {
      return null;
    });
  }

}

module.exports = Reader;
