'use strict';
const Excel = require('exceljs');
const fs = require('fs');
const path = require('path');

class Writer {

  constructor(config, workbook) {
    return new Promise(resolve => {
      const datas = config.datas;
      const total = config.total || 0;
      let singleMaxLimit = config.singleMaxLimit || 20000;
      const maxQuery = config.maxQuery || 20;
      let limit = 0;
      if (total > singleMaxLimit) {
        const pageSize = Math.ceil(total / singleMaxLimit);
        if (pageSize > maxQuery) {
          singleMaxLimit = Math.ceil(total / maxQuery);
        }
      }
      limit = singleMaxLimit;
      resolve(this.load(config, workbook, datas, total, limit, 0));
    });
  }

  push(config, workbook, datas) {
    const worksheet = workbook.getWorksheet(config.sheet_name);
    const headers = config.headers || [];
    const maxLength = datas.length;
    const unNumber = /\d+/g;

    for (let i = 0; i < datas.length; i++) {
      const node = datas[i];
      const data = [];
      const emptys = [];
      const names = [];
      const links = [];
      headers.forEach((target, index) => {
        const id = target.key;
        names.push(id);
        let value = node[id];
        if (target.formatter) {
          if (typeof (target.formatter) === 'function') {
            value = target.formatter(value, node);
          } else {
            value = value;
          }
        }
        const isEmpty = (!value || value == '' || value == null || value.length <= 0);
        if (isEmpty && target.empty) {
          value = target.empty;
          emptys.push(index);
        }
        if (target.link) {
          const link_sheet_name = target.link.sheet;
          const link_data = workbook.worksheetData[link_sheet_name];
          if (link_data) {
            const link_targets = target.link.targets(node, link_data);
            const link_target = target.link.link(node, link_targets);
            if (link_target) {
              links.push({
                index,
                sheet: link_sheet_name,
                key: link_target._map[target.link.key],
              });
            }
          }
        }
        const length = strlen(value);
        data[index] = value == null ? '' : value;
        const haeder_max = target.maxLength || 0;
        if (length > haeder_max) {
          target.maxLength = length;
          worksheet.getColumn(target.columnIndex).width = length + 2.62;
        }
      });
      const row = worksheet.addRow(data);
      const original = workbook.worksheetData[config.sheet_name];
      if (!original) {
        workbook.worksheetData[config.sheet_name] = [];
      }
      row.eachCell(function(cell, colNumber) {
        const name = names[colNumber - 1];
        const address = cell.address;
        if (!node._map) {
          node._map = {};
        }
        node._map[name] = address;
      });
      workbook.worksheetData[config.sheet_name].push(node);
      if (emptys.length > 0) {
        emptys.forEach(function(node, index) {
          row.getCell(node + 1).style.font = {
            color: {
              argb: 'FFE4E2E1',
            },
          };
        });
      }
      if (links.length > 0) {
        links.forEach(function(node, index) {
          const cell = row.getCell(node.index + 1);
          cell.style.font = {
            color: {
              argb: 'FF1278EA',
            },
          };
          cell.value = {
            formula: 'HYPERLINK("#' + node.sheet + '!' + node.key + '","1")',
            result: cell.value,
          };
        });
      }
      row.alignment = {
        vertical: 'middle',
        horizontal: 'center',
      };
      // datas.splice(0, 1);
      // datas.length++;
    }
    datas = null;
    return true;
  }

  create(config) {
    const workbook = this.createWorkbook(config);
    return workbook;
  }

  createWorbook(config = {}) {
    let workbook;
    if (config.isStream) {
      mkdirsSync('./cache/' + config.streamTarget);
      const options = {
        filename: './cache/' + config.streamTarget + '/' + config.filename + '.xlsx',
        useStyles: true,
        useSharedStrings: true,
      };
      workbook = new Excel.stream.xlsx.WorkbookWriter(options);
    } else {
      workbook = new Excel.Workbook();
    }
    workbook.creator = 'Ewsedu';
    workbook.lastModifiedBy = 'Ewsedu';
    workbook.created = new Date();
    workbook.modified = new Date();
    workbook.worksheetData = {};
    return workbook;
  }

  createWorksheet(config, workbook) {
    workbook = workbook || this.createWorkbook(config);
    config = config || {};
    const headers = config.headers || [];
    const groups = config.groups || null;
    const tops = groups;
    const datas = config.datas || [];
    const frozen = config.frozen || {};
    const mainTitle = config.title || {
      show: false,
      showTime: false,
      name: '导出数据',
    };
    const mainTime = config.time || {
      show: false,
    };
    mainTime.time = mainTime.time || new Date().toLocaleString();
    const line = 0;
    const unNumber = /\d+/g;

    const headbook = new Excel.Workbook();
    const views = this.createBasicSheet(config, headbook, null, false);
    const worksheet = this.createBasicSheet(config, workbook, views, true);
    return worksheet;
  }

  createBasicSheet(config, workbook, views, real) {
    views = views || [];
    const sheet_name = config.sheet_name || '导出数据';
    const headers = config.headers || [];
    const groups = Object.keys(config.groups) == 0 ? null : config.groups;
    const datas = config.datas || [];
    const frozen = config.frozen || {};
    const mainTitle = config.title || {
      show: false,
      showTime: false,
      name: '导出数据',
    };
    const mainTime = config.time || {
      show: false,
    };
    mainTime.time = mainTime.time || new Date().toLocaleString();
    let line = 0;
    const worksheet = workbook.addWorksheet(sheet_name, {
      views,
    });

    if (mainTitle.show) {
      line++;
      const titleRow = worksheet.addRow([ mainTitle.name ]);
      titleRow.height = 30;
      titleRow.alignment = {
        vertical: 'middle',
        horizontal: 'center',
      };
      titleRow.font = {
        name: 'Microsoft Yahei',
        family: 4,
        size: 16,
        bold: false,
      };
    }
    if (mainTime.show) {
      line++;
      worksheet.addRow([ '数据统计时间：' + mainTime.time ]);
    }
    const tops = groups;
    if (tops) {
      line++;
      const topRow = [];
      for (const k in tops) {
        topRow.push(tops[k]);
      }
      worksheet.addRow(topRow);
    }
    const row = [];
    let columnTarget = 0;
    headers.forEach(node => {
      ++columnTarget;
      node.columnIndex = columnTarget;
      row.push(node.name);
    });
    const unNumber = /\d+/g;
    const columnRow = worksheet.addRow(row);
    columnRow.alignment = {
      vertical: 'middle',
      horizontal: 'center',
    };
    let xSplit = 0;
    line++;

    config.columnIndex = line;

    columnRow.eachCell(function(cell, colNumber) {
      const head = headers.find(function(data) {
        return data.columnIndex == colNumber;
      });
      const column = cell.address.replace(unNumber, '');
      if (head) {
        head.column = column;
        head.address = cell.address;
        const length = strlen(head.name);
        head.maxLength = length;
        worksheet.getColumn(head.columnIndex).width = length + 2.62;
        if (head.frozen) {
          xSplit = colNumber;
        }
      }
    });

    xSplit = frozen.x ? xSplit : 0;
    const ySplit = frozen.y ? line : 0;
    const finalViews = [{
      state: 'frozen',
      xSplit,
      ySplit,
    }];
    const dataTop = line - 1;
    for (const k in tops) {
      const groups = headers.filter(data => {
        return data.group == k;
      });
      if (groups && groups.length > 0) {
        const begin = groups[0].address.replace(unNumber, '');
        const end = groups[groups.length - 1].address.replace(unNumber, '');
        worksheet.mergeCells(begin + dataTop + ':' + end + dataTop);
        const cell = worksheet.getCell(begin + dataTop);
        cell.value = tops[k];
        cell.alignment = {
          vertical: 'middle',
          horizontal: 'center',
        };
      }
    }
    const total_begin = headers[0].address.replace(unNumber, '');
    const total_end = headers[headers.length - 1].address.replace(unNumber, '');
    let titleTop = 0;
    if (mainTitle.show) {
      titleTop = 1;
      const title_merge = total_begin + titleTop + ':' + total_end + titleTop;
      worksheet.mergeCells(title_merge);
    }
    if (mainTime.show) {
      titleTop++;
      const date_merge = total_begin + titleTop + ':' + total_end + titleTop;
      worksheet.mergeCells(date_merge);
    }
    return real ? worksheet : finalViews;
  }

  export() {
    const loader = this.load();
    return loader.then(function(loaded) {
      const {
        config,
        workbook,
        datas,
      } = loaded;
      const worksheet = workbook.getWorksheet(1);
      if (config.isStream) {
        return workbook.commit().then(function() {
          return {
            type: 'file',
            data: '/files/cache/' + config.streamTarget + '/' + config.filename + '.xlsx',
          };
        });
      }
      return workbook.xlsx.writeBuffer().then(function(buffer) {
        return {
          type: 'buffer',
          data: buffer,
        };
      });

    }).catch(function(e) {
      return e.message;
    });
  }

}

function mkdirsSync(dirname, mode) {
  if (fs.existsSync(dirname)) {
    return true;
  }
  if (mkdirsSync(path.dirname(dirname), mode)) {
    fs.mkdirSync(dirname, mode);
    return true;
  }

}
function getQueryParam(url, key) {
  const parse = urlMaster.parse_url(url);
  const query = parse.query || '';
  const queries = query.split('&');
  const output = {};
  queries.forEach(function(node, index) {
    const info = node.split('=');
    if (info[0]) {
      output[info[0]] = info[1];
    }
  });
  return key ? output[key] : output;
}

function strlen(str) {
  if (str == null) return 0;
  if (typeof str !== 'string') {
    str += '';
  }
  return str.replace(/[^\x00-\xff]/g, '01').length;
}

module.exports = Writer;
